//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "GameMap.h"
#include <iostream>
#include <sstream>
#include <fstream>
#include "citylist.h"
#include "city.h"
#include "ruin.h"
#include "temple.h"
#include "playerlist.h"
#include "stacklist.h"

using namespace std;

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<std::endl<<flush;}
#define debug(x)

GameMap* GameMap::s_instance = 0;

int GameMap::s_size = 100;

GameMap* GameMap::getInstance()
{
    if (s_instance != 0)
      return s_instance;
    else return 0;
}


GameMap* GameMap::getInstance(std::string TilesetName)
{
    if (s_instance == 0)
    {
        s_instance = new GameMap(TilesetName);

    }
    return s_instance;
}

GameMap* GameMap::getInstance(XML_Helper* helper)
{
    if (s_instance)
        deleteInstance();

    s_instance = new GameMap(helper);

    return s_instance;
}

void GameMap::deleteInstance()
{
    if (s_instance)
        delete s_instance;
    s_instance = 0;
}

GameMap::GameMap(std::string TilesetName)
{
    d_tileSet = new TileSet(TilesetName);

    d_map = new Maptile*[s_size*s_size];
    for (int j = 0; j < s_size; j++)
        for (int i = 0; i < s_size; i++)
            d_map[j*s_size + i] = 0;

}

GameMap::GameMap(XML_Helper* helper)
{
    std::string types;
    std::string t_name;

    helper->getData(s_size, "size");
    helper->getData(t_name,"tileset", false);
    helper->getData(types, "types", false);

    d_tileSet = new TileSet(t_name);

    //create the map
    d_map = new Maptile*[s_size*s_size];

    int offset = 0;
    for (int j = 0; j < s_size; j++)
    {
        // remove newline and carriage return lines
        char test = types[j*s_size + offset];
        while (test == '\n' || test == '\r')
        {
            offset++;
            test = types[j*s_size + offset];
        }

        for (int i = 0; i < s_size; i++)
        {
            //due to the circumstances, types is a long stream of
            //numbers, so read it character for character (no \n's or so)
            char type = types[j*s_size + i + offset];  

            //the chars now hold the ascii representation of the numbers, which
            //we don't want
            type -= '0';
            d_map[j*s_size + i] = new Maptile(d_tileSet, i, j, (*d_tileSet)[type]->getType());
        }
    }

    //add some callbacks for item loading
    helper->registerTag("itemstack", SigC::slot(*this, &GameMap::loadItems));
    helper->registerTag("item", SigC::slot(*this, &GameMap::loadItems));
}


GameMap::~GameMap()
{
    delete d_tileSet;

    for (int i = 0; i < s_size; i++)
    {
        for (int j = 0; j < s_size; j++)
        {
            if (d_map[j*s_size + i])
                delete d_map[j*s_size + i];
        }
    }

    delete[] d_map;
}

bool GameMap::fill(MapGenerator* generator)
//basically, this does the same as the former random function, but you don't
//need to go the whole way via dumping the map in a file etc.
{
    int size;
    const Tile::Type* terrain = generator->getMap(size);


    //the sizes should definitely match, else we have a problem here
    if (size != s_size)
    {
        std::cerr <<_("Error in GameMap::fillMap: sizes don't match!! Exiting.\n");
        exit(-1);
    }

    // create tiles and set their buildings accordingly
    for (int j = 0; j < size; j++)
        for (int i = 0; i < size; i++)
            d_map[j*s_size + i] = new Maptile(d_tileSet, i, j, terrain[j*size + i]);

    return true;
}

bool GameMap::fill(Tile::Type type)
{
    for (int i = 0; i < s_size; i++)
        for (int j = 0; j < s_size; j++)
            d_map[j*s_size + i] = new Maptile(d_tileSet, i, j, type);

    return true;
}

bool GameMap::save(XML_Helper* helper) const
{
    bool retval = true;

    std::stringstream types;

    types <<endl;
    for (int i = 0; i < s_size; i++)
    {
        for (int j = 0; j < s_size; j++)
            types << getTile(j, i)->getType();
        types <<endl;
    }


    retval &= helper->openTag("map");
    retval &= helper->saveData("size", s_size);
    retval &= helper->saveData("tileset", d_tileSet->getName());
    retval &= helper->saveData("types", types.str());

    // last, save all items lying around somewhere
    for (int i = 0; i < s_size; i++)
        for (int j = 0; j < s_size; j++)
            if (!getTile(i,j)->getItems().empty())
            {
                retval &= helper->openTag("itemstack");
                retval &= helper->saveData("x", i);
                retval &= helper->saveData("y", j);
                
                std::list<Item*> items = getTile(i,j)->getItems();
                
                std::list<Item*>::const_iterator it;
                for (it = items.begin(); it != items.end(); it++)
                    (*it)->save(helper);

                retval &= helper->closeTag();
            }
     
    retval &= helper->closeTag();
    return retval;
}

bool GameMap::loadItems(std::string tag, XML_Helper* helper)
{
    static int x = 0;
    static int y = 0;
    
    if (tag == "itemstack")
    {
        helper->getData(x, "x");
        helper->getData(y, "y");
    }

    if (tag == "item")
    {
        Item* item = new Item(helper);
        getTile(x, y)->addItem(item);
    }

    return true;
}

Maptile* GameMap::getTile(int x, int y) const
{
    if ((x < 0) || (x >= s_size) || (y < 0) || (y >= s_size))
    {
        std::cerr <<"GameMap::getTile: function tried to access tile out of bonds at (";
        std::cerr <<x <<"," <<y <<")! Exiting\n" <<std::flush;

        exit(-1);
    }

    return d_map[y*s_size + x];
}

// End of file
