//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "GraphicsCache.h"
#include "armysetlist.h"
#include "army.h"
#include "playerlist.h"
#include "Configuration.h"
#include "File.h"
#include "GameMap.h"

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<std::endl<<std::flush;}
#define debug(x)

//some structures first

//this structure is the base class for storing cached objects. It stores the
//(army-, but can be extended) set, the type of army and the player which the
//surface is designed for (and the surface itself, of course).
struct ArmyCacheItem
{
    Uint32 armyset;
    Uint32 index;
    const Player* player;
    int level;
    bool medals[3];
    SDL_Surface* surface;
};

//the structure to store buildings in
struct CityCacheItem
{
    int type;
    const Player* player;
    SDL_Surface* surface;
};

// the structure to store flags in
struct FlagCacheItem
{
    Uint32 size;
    const Player* player;
    SDL_Surface* surface;
};

//-----------------------------------------------------

GraphicsCache* GraphicsCache::s_instance = 0;

GraphicsCache* GraphicsCache::getInstance()
{
    if (!s_instance)
        s_instance = new GraphicsCache();

    return s_instance;
}

void GraphicsCache::deleteInstance()
{
    if (!s_instance)
        return;

    delete s_instance;
    s_instance =0;
}

GraphicsCache::GraphicsCache()
    :d_cachesize(0)
{
    loadCityPics();
    loadFlags();

    d_levelmask = File::getMiscPicture("level_mask.png");
    d_medalsmask = File::getMiscPicture("medals_mask.gif");
}

GraphicsCache::~GraphicsCache()
{
    clear();

    for (int i = 0; i < 3; i++)
    {
        if (d_citypic[i])
            SDL_FreeSurface(d_citypic[i]);

        if (d_citymask[i])
            SDL_FreeSurface(d_citymask[i]);
    }

    for (int i = 0; i < 8; i++)
    {
        SDL_FreeSurface(d_flagpic[i]);
        SDL_FreeSurface(d_flagmask[i]);
    }

    SDL_FreeSurface(d_razedpic);
    SDL_FreeSurface(d_levelmask);
    SDL_FreeSurface(d_medalsmask);
}

SDL_Surface* GraphicsCache::getArmyPic(Uint32 armyset, Uint32 army, const Player* p,
                                       int level, const bool *medals)
{
    debug("getting army pic " <<armyset <<" " <<army <<" " <<p->getName())

    std::list<ArmyCacheItem*>::iterator it;
    ArmyCacheItem* myitem;

    // important: medals can be 0 in special cases
    bool my_medals[3] = {0,0,0};
    if (medals)
        for (int i = 0; i < 3; i++)
            my_medals[i] = medals[i];

    // special situation: ruin keepers don't belong to any player
    if (!p)
        p = Playerlist::getInstance()->getNeutral();
    
    for (it =d_armylist.begin(); it != d_armylist.end(); it++)
    {
        if (((*it)->armyset == armyset) && ((*it)->index == army)
            && ((*it)->player ==p) && ((*it)->level==level) 
            && ((*it)->medals[0]==my_medals[0])
            && ((*it)->medals[1]==my_medals[1])
            && ((*it)->medals[2]==my_medals[2]))
        {
            myitem = (*it);
            
            // put the item on the last place (==last touched)
            d_armylist.erase(it);
            d_armylist.push_back(myitem);
            
            return myitem->surface;
        }
    }

    // We are still here, so the graphic is not in the cache. addArmyPic calls
    // checkPictures on its own, so we can simply return the surface
    debug("getarmypic============= " << my_medals) 
    myitem = addArmyPic(armyset, army, p, level, my_medals);

    return myitem->surface;
}

SDL_Surface* GraphicsCache::getCityPic(const City* city)
{
    if (!city)
        return 0;

    if (city->isBurnt())
        return d_razedpic;

    return getCityPic(city->getDefenseLevel()-1, city->getPlayer());
}

SDL_Surface* GraphicsCache::getCityPic(int type, const Player* p)
{
    debug("GraphicsCache::getCityPic " <<type <<", player " <<p->getName())

    if (type == -1)
        return d_razedpic;
        
    if (!p)
    {
        std::cerr <<_("GraphicsCache: wanted to get city pic without player. Exiting...\n");
        exit(-1);
    }

    std::list<CityCacheItem*>::iterator it;
    CityCacheItem* myitem;

    for (it = d_citylist.begin(); it != d_citylist.end(); it++)
    {
        if (((*it)->type == type) && ((*it)->player == p))
        {
            myitem = (*it);

            //put the item in last place (last touched)
            d_citylist.erase(it);
            d_citylist.push_back(myitem);

            return myitem->surface;
        }
    }

    //no item found -> create a new one
    myitem = addCityPic(type, p);

    return myitem->surface;
}

SDL_Surface* GraphicsCache::getFlagPic(const Stack* s)
{
    debug("GraphicsCache::getFlagPic " <<s->getId() <<", player" <<s->getPlayer()->getName())

    if (!s)
    {
        std::cerr <<_("GraphicsCache::getFlagPic: no stack supplied! Exiting...\n");
        exit(-1);
    }
    
    std::list<FlagCacheItem*>::iterator it;
    FlagCacheItem* myitem;

    for (it = d_flaglist.begin(); it != d_flaglist.end(); it++)
    {
        myitem = *it;
        if ((myitem->size == s->size()) && (myitem->player == s->getPlayer()))
        {
            // put the item in last place (last touched)
            d_flaglist.erase(it);
            d_flaglist.push_back(myitem);

            return myitem->surface;
        }
    }

    // no item found => create a new one
    myitem = addFlagPic(s->size(), s->getPlayer());

    return myitem->surface;
}

void GraphicsCache::checkPictures()
{
    // for security, we always take a minimum cache size of 1MB. This
    // includes (at 4 byte color depth and assuming 50x50 pixel size)
    // - 10 cities (each 40kb => 400kb)
    // - 20 flags (each 10kb => 200kb)
    // - 40 units (each 10kb => 400kb)
    // + a bit more. The problem is that if we have less images than needed
    // for a single rendering, the surfaces will become invalid before actually
    // used. This should not be a problem with the normal map (the surfaces are
    // copied and discarded), but when selecting armies from armyset, where
    // you can have these images assigned to buttons.
    Uint32 maxcache = Configuration::s_cacheSize;
    if (maxcache < (1<<20))
        maxcache = (1<<20);
    
    if (d_cachesize < maxcache)
        return;
    

    // Now the cache size has been exceeded. We try to guarantee the values
    // given above and reduce the number of images. Let us start with the
    // cities
    
    while (d_citylist.size() > 10)
        eraseLastCityItem();

    // was this enough?
    if (d_cachesize < maxcache)
        return;

    // next, kill flag pics
    while (d_flaglist.size() > 20)
        eraseLastFlagItem();

    if (d_cachesize < maxcache)
        return;

    // still not enough? Erase army images
    while (d_armylist.size() > 40)
        eraseLastArmyItem();
}

ArmyCacheItem* GraphicsCache::addArmyPic(Uint32 armyset, Uint32 army,
                        const Player* p, int level, const bool *medalsbonus)
{
    debug("ADD army pic")

    ArmyCacheItem* myitem = new ArmyCacheItem();
    myitem->armyset = armyset;
    myitem->index = army;
    myitem->player = p;
    myitem->level = level;
    myitem->medals[0] = medalsbonus[0];
    myitem->medals[1] = medalsbonus[1];
    myitem->medals[2] = medalsbonus[2];

    //Now the most important part: load the army picture
    //First, copy the army picture and change it to the display format
    const Army* basearmy = Armysetlist::getInstance()->getArmy(armyset, army);
    myitem->surface = SDL_DisplayFormatAlpha(basearmy->getPixmap());

    //Now, if we have a mask ready, copy the mask over the picture. The mask
    //is <= 256 colors, we use the player's palette to colorize the mask. If
    //the armyset has no mask (happens during the transition phase), ignore this
    //step.
    if (basearmy->getMask())
    {
        SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
                            basearmy->getMask()->format->BitsPerPixel,0,0,0,0);

        //Set some color palette first!! Else we end up with a black surface.
        SDL_SetColors(mask, basearmy->getMask()->format->palette->colors, 0,
                            basearmy->getMask()->format->palette->ncolors);
                            
        SDL_BlitSurface(basearmy->getMask(), 0, mask, 0);

        //now change tmp's palette to the player's one
        SDL_SetColors(mask, p->getPalette(), 0, 16);

        //set the first pixel as alpha value
        SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

        //blit mask over the army pic
        SDL_BlitSurface(mask, 0, myitem->surface, 0);

        //free the temporary surface
        SDL_FreeSurface(mask);
    }
    
    if (level > 1 && d_levelmask)
    {
        SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
                              d_levelmask->format->BitsPerPixel,0,0,0,0);
        // a little hack while waiting for a complete level picture
        if (level<26) 
        {
            PG_Rect r(40*(level-2), 0, 40, 40);
            SDL_BlitSurface(d_levelmask, &r, mask, 0);
        }
        else 
        {
            PG_Rect r(160, 0, 40, 40);
            SDL_BlitSurface(d_levelmask, &r, mask, 0);
        }
      
        //set the first pixel as alpha value
        SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

        //blit mask over the army pic
        SDL_BlitSurface(mask,0, myitem->surface, 0);

        //free the temporary surface
        SDL_FreeSurface(mask);
    }

    if (d_medalsmask && medalsbonus != NULL)
    {
        debug("medalsbonus============= " << medalsbonus); 
        for(int i=0;i<3;i++)
        { 
            if (medalsbonus[i])
            {
                SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
                d_medalsmask->format->BitsPerPixel,0,0,0,0);

                // a little hack while waiting for a complete medals picture
                PG_Rect r(40*i, 0, 40, 40);
                SDL_BlitSurface(d_medalsmask, &r, mask, 0);
      
                //set the first pixel as alpha value
                SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

                //blit mask over the army pic
                SDL_BlitSurface(mask,0, myitem->surface, 0);

                //free the temporary surface
                SDL_FreeSurface(mask);
            }
        }
    }

    //now the final preparation steps:
    //a) add the size
    int size = myitem->surface->w * myitem->surface->h;
    d_cachesize += myitem->surface->format->BytesPerPixel * size;

    //b) add the entry to the list
    d_armylist.push_back(myitem);

    //c) check if the cache size is too large
    checkPictures();

    //we are finished, so return the pic
    return myitem;
}

CityCacheItem* GraphicsCache::addCityPic(int type, const Player* p)
{
    SDL_Surface* mysurf = SDL_DisplayFormatAlpha(d_citypic[type]);

    //now we do basically the same as in addArmyPic, just with other
    //pointers
    SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 100, 100,
                            d_citymask[type]->format->BitsPerPixel,0,0,0,0);

    //Set some color palette first!! Else we end up with a black surface.
    SDL_SetColors(mask, d_citymask[type]->format->palette->colors, 0,
                  d_citymask[type]->format->palette->ncolors);
                            
    SDL_BlitSurface(d_citymask[type], 0, mask, 0);

    //now change mask's palette to the player's one
    SDL_SetColors(mask, p->getPalette(), 0, 16);

    //set the first pixel as alpha value
    SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

    //blit mask over the army pic
    SDL_BlitSurface(mask, 0, mysurf, 0);

    //free the temporary surface
    SDL_FreeSurface(mask);

    //now create the cache item and add the size
    CityCacheItem* myitem = new CityCacheItem();
    myitem->player = p;
    myitem->type = type;
    myitem->surface = mysurf;

    d_citylist.push_back(myitem);

    //add the size
    int size = mysurf->w * mysurf->h;
    d_cachesize += size * mysurf->format->BytesPerPixel;

    //and check the size of the cache
    checkPictures();

    return myitem;
}

FlagCacheItem* GraphicsCache::addFlagPic(int size, const Player* p)
{
    debug("GraphicsCache::addFlagPic, player="<<p->getName()<<", size="<<size)
    
    // size is the size of the stack, but we need the index, which starts at 0
    size--;
    
    SDL_Surface* mysurf = SDL_DisplayFormatAlpha(d_flagpic[size]);

    //now we do basically the same as in addArmyPic, just with other
    //pointers
    SDL_Surface* mask = SDL_CreateRGBSurface(SDL_SWSURFACE, 100, 100,
                            d_flagmask[size]->format->BitsPerPixel,0,0,0,0);

    //Set some color palette first!! Else we end up with a black surface.
    SDL_SetColors(mask, d_flagmask[size]->format->palette->colors, 0,
                  d_flagmask[size]->format->palette->ncolors);
                            
    SDL_BlitSurface(d_flagmask[size], 0, mask, 0);

    //now change mask's palette to the player's one
    SDL_SetColors(mask, p->getPalette(), 0, 16);

    //set the first pixel as alpha value
    SDL_SetColorKey(mask, SDL_SRCCOLORKEY, 0);

    //blit mask over the army pic
    SDL_BlitSurface(mask, 0, mysurf, 0);

    //free the temporary surface
    SDL_FreeSurface(mask);

    //now create the cache item and add the size
    FlagCacheItem* myitem = new FlagCacheItem();
    myitem->player = p;
    myitem->size = size;
    myitem->surface = mysurf;

    d_flaglist.push_back(myitem);

    //add the size
    int picsize = mysurf->w * mysurf->h;
    d_cachesize += picsize * mysurf->format->BytesPerPixel;

    //and check the size of the cache
    checkPictures();

    return myitem;
}

void GraphicsCache::clear()
{
    while (!d_armylist.empty())
        eraseLastArmyItem();

    while (!d_citylist.empty())
        eraseLastCityItem();

    while (!d_flaglist.empty())
        eraseLastFlagItem();
}

void GraphicsCache::eraseLastArmyItem()
{
    if (d_armylist.empty())
        return;

    //As the name suggests, this function erases the last item in the list.
    //Whenever an item is requested, it moves to the first position, so the
    //last item is the oldest and therefore propably most useless in the list.
    ArmyCacheItem* myitem = *(d_armylist.begin());
    d_armylist.erase(d_armylist.begin());

    //don't forget to subtract the size from the size entry
    int size = myitem->surface->w * myitem->surface->h;
    d_cachesize -= myitem->surface->format->BytesPerPixel * size;

    SDL_FreeSurface(myitem->surface);
    delete myitem;
}

void GraphicsCache::eraseLastCityItem()
{
    if (d_citylist.empty())
        return;

    CityCacheItem* myitem = *(d_citylist.begin());
    d_citylist.erase(d_citylist.begin());

    int size = myitem->surface->w * myitem->surface->h;
    d_cachesize -= myitem->surface->format->BytesPerPixel * size;

    SDL_FreeSurface(myitem->surface);
    delete myitem;
}

void GraphicsCache::eraseLastFlagItem()
{
    if (d_flaglist.empty())
        return;

    FlagCacheItem* myitem = *(d_flaglist.begin());
    d_flaglist.erase(d_flaglist.begin());

    int size = myitem->surface->w * myitem->surface->h;
    d_cachesize -= myitem->surface->format->BytesPerPixel * size;

    SDL_FreeSurface(myitem->surface);
    delete myitem;
}

void GraphicsCache::loadCityPics()
{
    // GameMap has the actual tileset stored
    std::string tileset = GameMap::getInstance()->getTileSet()->getName();
   
    // load the image for the razed city
    d_razedpic = File::getMapsetPicture(tileset, "misc/castle_razed.png");
    
    // load the city pictures
    SDL_Surface* citypics = File::getMapsetPicture(tileset, "misc/castles.png");
    SDL_Surface* citymasks = File::getMapsetMask(tileset, "misc/castles_mask.png");

    // copy alpha values, don't use them
    SDL_SetAlpha(citypics, 0, 0);
    
    if (citymasks->format->BitsPerPixel > 8)
    {
        std::cout <<_("city masks have too high color depth, Exiting.\n");
        exit(-1);
    }
    
    for (int i = 0; i < 3; i++)
    {
        //copy the city image...
        SDL_Surface* tmp;
        SDL_PixelFormat* fmt = citypics->format;
        
        tmp = SDL_CreateRGBSurface(SDL_SWSURFACE, 100, 100, fmt->BitsPerPixel,
                            fmt->Rmask, fmt->Gmask, fmt->Bmask, fmt->Amask);

        PG_Rect r(i*100, 0, 100, 100);
        SDL_BlitSurface(citypics, &r, tmp, 0);

        d_citypic[i] = SDL_DisplayFormatAlpha(tmp);

        SDL_FreeSurface(tmp);
        
        //...and copy the mask image
        d_citymask[i] = SDL_CreateRGBSurface(SDL_HWSURFACE, 100, 100,
                        citymasks->format->BitsPerPixel,0,0,0,0);

        SDL_SetColors(d_citymask[i], citymasks->format->palette->colors, 0,
                        citymasks->format->palette->ncolors);
        PG_Rect r1(i*100, 0, 100, 100);
        SDL_BlitSurface(citymasks, &r1, d_citymask[i], 0);
    }

    SDL_FreeSurface(citypics);
    SDL_FreeSurface(citymasks);
}

void GraphicsCache::loadFlags()
{
    //GameMap has the actual tileset stored
    std::string tileset = GameMap::getInstance()->getTileSet()->getName();
    
    // to build flags, we need these three images as basic blocks
    SDL_Surface* flag = File::getMapsetPicture(tileset, "misc/flag.png");
    SDL_Surface* flagmask = File::getMapsetMask(tileset, "misc/flag_mask.png");
    SDL_PixelFormat* fmt = flag->format;
    
    if (flagmask->format->BitsPerPixel > 8)
    {
        std::cout <<_("flag masks have too high color depth. Exiting!\n");
        exit(-1);
    }
    
    // copy alpha values, don't use them!
    SDL_SetAlpha(flag, 0, 0);
    

    for (int i = 0; i < 8; i++)
    {
        // first, create the flag image
        SDL_Surface* tmp = SDL_CreateRGBSurface(SDL_SWSURFACE, 50, 50, 
                                    fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
                                    fmt->Bmask, fmt->Amask);
        
        // now blit the correct flag on the top of the image
        PG_Rect flagrect(i*50, 0, 50, 50);
        SDL_BlitSurface(flag, &flagrect, tmp, 0);

        // convert the surface to screen resolution etc.
        d_flagpic[i] = SDL_DisplayFormatAlpha(tmp);

        // free the temporary surface
        SDL_FreeSurface(tmp);


        // now create the masks
        d_flagmask[i]=  SDL_CreateRGBSurface(SDL_SWSURFACE, 50, 50,
                            flagmask->format->BitsPerPixel,0,0,0,0);
        SDL_SetColors(d_flagmask[i], flagmask->format->palette->colors, 0,
                        flagmask->format->palette->ncolors);
        SDL_BlitSurface(flagmask, &flagrect, d_flagmask[i], 0);
    }

    // free the temporary surfaces
    SDL_FreeSurface(flag);
    SDL_FreeSurface(flagmask);
}
