//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <iostream>
#include <expat.h>
#include <SDL_image.h>
#include <SDL.h>

#include "armysetlist.h"
#include "File.h"
#include "defs.h"


using namespace std;

#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl<<flush;}
//#define debug(x)

Armysetlist* Armysetlist::s_instance = 0;

Armysetlist* Armysetlist::getInstance()
{
    if (!s_instance)
        s_instance = new Armysetlist();

    return s_instance;
}

void Armysetlist::deleteInstance()
{
    if (s_instance)
        delete s_instance;

    s_instance = 0;
}

Armysetlist::Armysetlist()
    :d_standard(1), d_heroes(2), d_loading(0)
{
    // load all armysets
    std::list<std::string> armysets = File::scanArmysets();

    std::list<std::string>::const_iterator it;
    for (it = armysets.begin(); it != armysets.end(); it++)
        loadArmyset(*it);
}

Armysetlist::~Armysetlist()
{
    // remove all army entries
    ArmyMap::iterator it;
    for (it = d_armies.begin(); it != d_armies.end(); it++)
        while (!(*it).second.empty())
            delete ((*it).second)[0];
}

const Army* Armysetlist::getArmy(Uint32 id, Uint32 index) const
{
    // always use ArmyMap::find for searching, else a default entry is created,
    // which can produce really bad results!!
    ArmyMap::const_iterator it = d_armies.find(id);

    // armyset does not exist
    if (it == d_armies.end())
        return 0;

    // index too large
    if (index >= (*it).second.size())
        return 0;

    return ((*it).second)[index];
}

Uint32 Armysetlist::getSize(Uint32 id) const
{
    ArmyMap::const_iterator it = d_armies.find(id);

    // armyset does not exist
    if (it == d_armies.end())
        return 0;

    return (*it).second.size();
}

std::string Armysetlist::getName(Uint32 id) const
{
    NameMap::const_iterator it = d_names.find(id);

    // armyset does not exist
    if (it == d_names.end())
        return 0;

    return (*it).second;
}

std::vector<Uint32> Armysetlist::getArmysets(bool force_all) const
{
    std::vector<Uint32> retlist;
    
    NameMap::const_iterator it;
    for (it = d_names.begin(); it != d_names.end(); it++)
    {
        if (!force_all && ((*it).first == d_standard || (*it).first == d_heroes))
            continue;
        retlist.push_back((*it).first);
    }

    return retlist;
}

bool Armysetlist::loadArmyset(std::string name)
{
    debug("Loading armyset " <<name)
    
    // first, load the description file; loadGlobalStuff will care for setting
    // up d_loading (the id of the currently loaded armyset) and an entry in
    // the lists.
    XML_Helper helper(File::getArmyset(name), ios::in, false);
    helper.registerTag("armyset", SigC::slot((*this), &Armysetlist::loadGlobalStuff));
    helper.registerTag("army", SigC::slot((*this), &Armysetlist::loadArmy));
    
    if (!helper.parse())
    {
        std::cerr <<_("Error, while loading an armyset. Armyset Name: ");
        std::cerr <<name <<std::endl <<std::flush;
        exit(-1);
    }


    // Second, load the army images. The image file is just a long stripe of
    // all army images.
    SDL_Surface* armyGallery = File::getArmyPicture(name);
    SDL_PixelFormat* fmt = armyGallery->format;
    
    // don't use alpha information, just copy the channel! very important
    SDL_SetAlpha(armyGallery, 0, 0);
    
    for (unsigned int i = 0; i < d_armies[d_loading].size(); i++)
    {
        SDL_Surface* tmp = SDL_CreateRGBSurface(SDL_SWSURFACE,
                40, 40, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
                fmt->Bmask, fmt->Amask);

        PG_Rect r(40*i, 0, 40, 40);
        SDL_BlitSurface(armyGallery, &r, tmp, 0);

        SDL_Surface* pixmap = SDL_DisplayFormatAlpha(tmp);
        SDL_FreeSurface(tmp);
        (d_armies[d_loading])[i]->setPixmap(pixmap);
    }
    SDL_FreeSurface(armyGallery);
   
   
    // Third, load the masks
    armyGallery = File::getArmyMask(name);
    
    // no mask file => ignore it
    if (!armyGallery)
        return true;
    
    fmt = armyGallery->format;

    //The mask surface should _always_ have a 256 or lesser color palette. If
    //this isn't the case, the mask pictures are not valid. See GraphicsCache 
    //for more details how they are applied.
    if (fmt->BitsPerPixel > 8)
        return true;

    for (unsigned int i = 0; i < d_armies[d_loading].size(); i++)
    {
        //most of the format information is simply not needed
        SDL_Surface* maskimage;
        maskimage = SDL_CreateRGBSurface(SDL_SWSURFACE, 40, 40,
                                    fmt->BitsPerPixel, 0, 0, 0, 0);

        //since we have 256 color pics, set the palette, else we end up with
        //a monochromatic surface!!
        SDL_SetColors(maskimage, fmt->palette->colors,0, fmt->palette->ncolors);
        
        PG_Rect r(40*i, 0, 40, 40);
        SDL_BlitSurface(armyGallery, &r, maskimage, 0);

        (d_armies[d_loading])[i]->setMask(maskimage);
    }
    SDL_FreeSurface(armyGallery);
    
    return true;
}

bool Armysetlist::loadGlobalStuff(std::string tag, XML_Helper* helper)
{
    bool retval = true;
    std::string name;
    
    // load the data
    retval &= helper->getData(d_loading, "id");
    retval &= helper->getData(name, "name");

    // create the neccessary entry in the name list
    d_names[d_loading] = name;
    
    return retval;
}

bool Armysetlist::loadArmy(string tag, XML_Helper* helper)
{
    Army* a = new Army(helper, true);
    a->setArmyset(d_loading, d_armies[d_loading].size());

    d_armies[d_loading].push_back(a);

    return true;
}

