//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <algorithm>
#include <fstream>
#include <sstream>

#include "player.h"
#include "playerlist.h"
#include "FightDialog.h"
#include "stacklist.h"
#include "citylist.h"
#include "city.h"
#include "path.h"
#include "GameMap.h"
#include "armysetlist.h"
#include "real_player.h"
#include "ai_dummy.h"
#include "ai_fast.h"
#include "ai_smart.h"
#include "counter.h"
#include "army.h"
#include "AI_Analysis.h"
#include "AI_Allocation.h"

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<flush<<endl;}
#define debug(x)

// signal
SigC::Signal1<void, Player::Type> sendingTurn;

Player::Player(string name, Uint32 armyset, SDL_Color color, Type type)
    :d_color(color), d_name(name), d_armyset(armyset), d_gold(1000),
    d_dead(false), d_immortal(false), d_type(type)
{
    d_id= fl_counter->getNextId();
    d_stacklist = new Stacklist();
    debug("type of " << d_name << " is " << type)
        
    //init the palette
    d_palette = new SDL_Color[16];
    initPalette();

    d_fogmap = new FogMap();
}

Player::Player(const Player& player)
    :d_color(player.d_color), d_name(player.d_name), d_armyset(player.d_armyset),
    d_gold(player.d_gold), d_dead(player.d_dead), d_immortal(player.d_immortal),
    d_type(player.d_type), d_id(player.d_id), d_fogmap(player.d_fogmap)
{
    //d_palette is the 16 color palette associated with each player. For a
    //more exhaustive explanation, see Player::initPalette()
    d_palette = new SDL_Color[16];
    initPalette();

    // as the other player is propably dumped somehow, we need to deep copy
    // everything. This costs a lot, but the only useful situation for this
    // I can think of is a change of the player type, as occurs in the editor.
    Stacklist::iterator it;
    d_stacklist = new Stacklist();
    for (it = player.d_stacklist->begin(); it != player.d_stacklist->end(); it++)
    {
        Stack* mine = new Stack(**it);
        // change the stack's loyalty
        mine->setPlayer(this);
        d_stacklist->push_back(mine);
    }

    // copy actions
    std::list<Action*>::const_iterator ait;
    for (ait = player.d_actions.begin(); ait != player.d_actions.end(); ait++)
        d_actions.push_back(Action::copy(*ait));

    // copy fogmap; TBD
    d_fogmap = new FogMap();
}

Player::Player(XML_Helper* helper)
{
    helper->getData(d_id, "id");
    helper->getData(d_name, "name");
    helper->getData(d_gold, "gold");
    helper->getData(d_dead, "dead");
    helper->getData(d_immortal, "immortal");
    helper->getData(d_type, "type");

    string s;
    helper->getData(s, "color", false);
    
    int i;
    istringstream scolor(s);
    scolor >> i; d_color.r = i;
    scolor >> i; d_color.g = i;
    scolor >> i; d_color.b = i;

    helper->getData(d_armyset, "armyset");

    //last but not least, register the load function for actionlist
    helper->registerTag("action", SigC::slot((*this), &Player::load));
    helper->registerTag("stacklist", SigC::slot((*this), &Player::load));
    helper->registerTag("fogmap", SigC::slot((*this), &Player::load));

    //init the player's palette
    d_palette = new SDL_Color[16];
    initPalette();
}

Player::~Player()
{
    d_stacklist->flClear();
    delete d_stacklist;
    delete[] d_palette;
    delete d_fogmap;

    for (std::list<Action*>::iterator it = d_actions.begin(); it != d_actions.end(); it++)
    {
        delete (*it);
    }
}

Player* Player::create(std::string name, Uint32 armyset, SDL_Color color, Type type)
{
    switch(type)
    {
        case HUMAN:
            return new RealPlayer(name, armyset, color);
        case AI_FAST:
            return new AI_Fast(name, armyset, color);
        case AI_DUMMY:
            return new AI_Dummy(name, armyset, color);
        case AI_SMART:
            return new AI_Smart(name, armyset, color);
    }

    return 0;
}

Player* Player::create(Player* orig, Type type)
{
    switch(type)
    {
        case HUMAN:
            return new RealPlayer(*orig);
        case AI_FAST:
            return new AI_Fast(*orig);
        case AI_DUMMY:
            return new AI_Dummy(*orig);
        case AI_SMART:
            return new AI_Smart(*orig);
    }

    return 0;
}

void Player::setColor(SDL_Color c)
{
    d_color = c;
    initPalette();
}

void Player::addGold(int gold)
{
    d_gold += gold;
    schangingStatus.emit();
}

void Player::withdrawGold(int gold)
{
    d_gold -= gold;
    schangingStatus.emit();
}

Stack* Player::getActivestack()
{
    return d_stacklist->getActivestack();
}

void Player::initPalette()
{
    //Each player has, beside his color, a 16 color palette derived from his
    //player color. This palette is then assigned to "unit mask" surfaces, which
    //are blitted over the units to give them player-specific colors. The colors
    //in the palette are:
    //#0 dummy (is the alpha value of the mask)
    //#1 color for writing on dark backgrounds
    //#2 for writing on bright backgrounds
    //#3 - #8 bright player colors (#3 == brightest)
    //#9 player color
    //#10 - #15 darker player colors (#15 == darkest)

    //#0 is unimportant. To guarantee that it doesn't cover other colors, I set
    //it to pure white.
    d_palette[0].r = 255;
    d_palette[0].g = 255;
    d_palette[0].b = 255;

    //Curently, I don't see a use for writing, so #1 and #2 are just the player
    //color
    d_palette[1] = d_color;
    d_palette[2] = d_color;

    //#9 is the player color as well
    d_palette[9] = d_color;

    //Now the bright colors. I try a percentage approach (can be improved later
    //on). Each color portion (red/green/blue) has some difference to 255. This
    //difference is reduced by ten percent for each brighter color. E.g. if
    //r=140, then dr = 255-140 = 115. Color #8 will therefore have a difference
    //of 115 * (1 - 1/10) = 103 to the maximum. Thus, the red color part of
    //color #8 would be 255 - 103 = 152.
    int diffred, diffgreen, diffblue;
    SDL_Color tempcol;
    tempcol.unused=0;

    diffred = (255 - d_color.r) / 10;
    diffgreen = (255 - d_color.g) / 10;
    diffblue = (255 - d_color.b) / 10;

    for (int i = 1; i <= 6; i++)
    {
        tempcol.r = d_color.r + (diffred * i);
        tempcol.g = d_color.g + (diffgreen * i);
        tempcol.b = d_color.b + (diffblue * i);

        d_palette[9-i] = tempcol;
    }

    //The same procedure takes place with the darker colors.
    diffred = d_color.r / 10;
    diffgreen = d_color.g / 10;
    diffblue = d_color.b / 10;

    for (int i = 1; i <= 6; i++)
    {
        tempcol.r = d_color.r - (diffred * i);
        tempcol.g = d_color.g - (diffgreen * i);
        tempcol.b = d_color.b - (diffblue * i);

        d_palette[9+i] = tempcol;
    }

    //finished...
}

void Player::dumpActionlist() const
{
    for (list<Action*>::const_iterator it = d_actions.begin();
        it != d_actions.end(); it++)
    {
        cerr <<(*it)->dump() << endl;
    }    
}

void Player::clearActionlist()
{
    for (list<Action*>::iterator it = d_actions.begin();
        it != d_actions.end(); it++)
    {
        delete (*it);
    }
    d_actions.clear();
}

void Player::addStack(Stack* stack)
{
    stack->setPlayer(this);
    d_stacklist->push_back(stack);
}

bool Player::deleteStack(Stack* stack)
{
    AI_Analysis::deleteStack(stack);
    AI_Allocation::deleteStack(stack);
    return d_stacklist->flRemove(stack);
}

void Player::kill()
{
    if (d_immortal)
        // ignore it
        return;
    
    d_dead = true;
    d_stacklist->flClear();

    // Since in some cases the player can be killed rather innocently
    // (using reactions), we also need to clear the player's traces in the
    // single cities
    Citylist* cl = Citylist::getInstance();
    for (Citylist::iterator it = cl->begin(); it != cl->end(); it++)
        if ((*it).getPlayer() == this)
            (*it).setPlayer(Playerlist::getInstance()->getNeutral());
}

bool Player::save(XML_Helper* helper) const
{
    bool retval = true;

    // we do not want to have the character of the colors written out
    // (savefile is unicode encoded => may create problems)
    std::stringstream s;
    s << static_cast<Uint32>(d_color.r) <<" ";
    s << static_cast<Uint32>(d_color.g) <<" ";
    s << static_cast<Uint32>(d_color.b);

    retval &= helper->saveData("id", d_id);
    retval &= helper->saveData("name", d_name);
    retval &= helper->saveData("color", s.str());
    retval &= helper->saveData("armyset", d_armyset);
    retval &= helper->saveData("gold", d_gold);
    retval &= helper->saveData("dead", d_dead);
    retval &= helper->saveData("immortal", d_immortal);
    retval &= helper->saveData("type", d_type);
    debug("type of " << d_name << " is " << d_type)

    //save the actionlist
    for (list<Action*>::const_iterator it = d_actions.begin();
            it != d_actions.end(); it++)
        retval &= (*it)->save(helper);

    retval &= d_stacklist->save(helper);
    retval &= d_fogmap->save(helper);

    return retval;
}

Player* Player::loadPlayer(XML_Helper* helper)
{
    Type type;
    int t;
    
    helper->getData(t, "type");
    type = static_cast<Type>(t);

    switch (type)
    {
        case HUMAN:
            return (new RealPlayer(helper));
        case AI_FAST:
            return (new AI_Fast(helper));
        case AI_SMART:
            return (new AI_Smart(helper));
        case AI_DUMMY:
            return (new AI_Dummy(helper));
    }

    return 0;
}

bool Player::load(string tag, XML_Helper* helper)
{
    if (tag == "action")
    {
        Action* action;
        action = Action::handle_load(helper);
        d_actions.push_back(action);
    }

    if (tag == "stacklist")
        d_stacklist = new Stacklist(helper);

    if (tag == "fogmap")
        d_fogmap = new FogMap(helper);

    return true;
}

// End of file
