//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <iostream>
#include <pgdraw.h>
#include <pgapplication.h>
#include "vectormap.h"
#include "stacklist.h"
#include "citylist.h"
#include "ruinlist.h"
#include "templelist.h"
#include "city.h"
#include "stack.h"
#include "armysetlist.h"
#include "army.h"
#include "GameMap.h"
#include "path.h"
#include "ruin.h"
#include "temple.h"
#include "playerlist.h"
#include "player.h"
#include "config.h"

using namespace std;

#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl<<flush;}
//#define debug(x)

VectorMap::VectorMap(City * city, PG_Widget* parent, PG_Rect rect, int pixelsize)
  :PG_Widget(parent, rect, true), d_city(city),d_pixels(pixelsize)
{

    SDL_PixelFormat* fmt = PG_Application::GetScreen()->format;
    d_staticMap = SDL_CreateRGBSurface(SDL_SWSURFACE, GameMap::getSize()*d_pixels,
            GameMap::getSize()*d_pixels, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
            fmt->Bmask, fmt->Amask);
    
    // Draw map
    SDL_Color c;
    for(int i = 0; i < GameMap::getSize(); i++)
    {
        for(int j = 0; j < GameMap::getSize(); j++)
        {
            c = GameMap::getInstance()->getTile(i,j)->getColor();

            for (int dx = 0; dx < d_pixels; dx++)
                for (int dy = 0; dy < d_pixels; dy++)
                    PG_Draw::SetPixel(i*d_pixels + dx, j*d_pixels + dy,
                                      PG_Color(c.r, c.g, c.b), d_staticMap);
        }
    }
}

VectorMap::~VectorMap()
{
    SDL_FreeSurface(d_staticMap);
}

void VectorMap::eventDraw(SDL_Surface* surface, const PG_Rect& rect)
{
    debug("eventDraw()");

    PG_Rect r(0, 0, GameMap::getSize() * d_pixels, GameMap::getSize() * d_pixels);
    SDL_BlitSurface(d_staticMap, 0, surface, &r);

    // Draw all cities as rectangles around the city location. They consist
    // of an outer rectangle in black and inner ones in player colors.
    for (Citylist::iterator it = Citylist::getInstance()->begin();
        it != Citylist::getInstance()->end(); it++)
    {    
        SDL_Color c = (*it).getPlayer()->getColor();
        PG_Point pos = (*it).getPos();
        
        if (&(*it) == d_city)
            DrawRectWH(pos.x*d_pixels, pos.y*d_pixels, 2*d_pixels, 2*d_pixels,PG_Color(255, 0, 255)); 
        else
            DrawRectWH(pos.x*d_pixels, pos.y*d_pixels, 2*d_pixels, 2*d_pixels,PG_Color(0, 0, 0)); 

        for (int i=1;i<2*d_pixels;i++) 
            DrawRectWH(pos.x*d_pixels+i, pos.y*d_pixels+i, 2*d_pixels-(i*2), 2*d_pixels-(i*2),PG_Color(c));
    }


    // Draw the vector
    if (d_city->getVectoring().x != -1)
    {
        debug("Draw vector")
        PG_Point start = d_city->getPos();
        PG_Point end = d_city->getVectoring();
        
        DrawLine(start.x*d_pixels, start.y*d_pixels, end.x*d_pixels, end.y*d_pixels, PG_Color(0,0,0));
    }

    DrawBorder(PG_Rect(0, 0, my_width, my_height), 1);
}

bool VectorMap::eventMouseButtonDown(const SDL_MouseButtonEvent* event)
{  

    // actualize the mouse position
    PG_Point oldpos = d_pos;
    d_pos.x = event->x/d_pixels - my_xpos/d_pixels;
    d_pos.y = event->y/d_pixels - my_ypos/d_pixels;

    switch (event->button)
    {
        case SDL_BUTTON_LEFT:
	{

           // Here we must check if the army produced by the city can be sent to the vector location
           // so we create a temporary stack and the fill it with an army that is the current production of the city.
           // Then calculate if exist a plausible path that brings from the city to the vector location.
           // If the path exists then we set the city vectoring else we unset the vectoring.

            PG_Point pos=d_city->getPos();
            Stack* st = new Stack(Playerlist::getActiveplayer(),pos);

            debug("Vectoring from =" << pos.x << "," << pos.y)

            const Armysetlist* al = Armysetlist::getInstance();
            Uint32 set;
            int index=0;
            int val=0;

            if (d_city->getAdvancedProd())
            {
                set = d_city->getPlayer()->getArmyset();
                index = d_city->getArmytype(d_city->getProductionIndex(),true);
            }
            else
            {
                set = al->getStandardId();
                index = d_city->getArmytype(d_city->getProductionIndex(),false);
            }
        
            // The city can have no production so we calculate the path only when index != 0
            if (index!= -1) 
            { 
                debug("Index=" << index)

                st->push_back(new Army(*(al->getArmy(set, index)), d_city->getPlayer()));
                val= st->getPath()->calculate(st, d_pos);
         
                debug("Vectoring to   =" << d_pos.x << "," << d_pos.y)
                debug("Path Value=" << val)
	    }

            if (val!=0 && index!=-1)
            {
                sclickVectMouse.emit(PG_Point(d_pos.x,d_pos.y));
                d_city->setVectoring(d_pos);
            }
            else   
            {
                sclickVectMouse.emit(PG_Point(-1,-1));
                d_city->setVectoring(PG_Point(-1,-1));
            }

	    delete st;
            break;
	}
        case SDL_BUTTON_RIGHT:
	{
              
	    // We use the right button to unset the city vectoring 
            sclickVectMouse.emit(PG_Point(-1,-1));
            d_city->setVectoring(PG_Point(-1,-1));
            break;
	}
        default:
            break;
    }

    Redraw();
    return true;
}

bool VectorMap::eventMouseMotion(const SDL_MouseMotionEvent* event)
{
    // actualize the mouse position
    PG_Point oldpos = d_pos;
    d_pos.x = event->x/d_pixels - my_xpos/d_pixels;
    d_pos.y = event->y/d_pixels - my_ypos/d_pixels;

    int diff_xtiles = d_pos.x - oldpos.x;
    int diff_ytiles = d_pos.y - oldpos.y;

    if ((diff_xtiles == 0) && (diff_ytiles == 0))
        return true;

    smovVectMouse.emit(PG_Point(d_pos.x,d_pos.y));
    
    return true;
}

void VectorMap::eventMouseLeave()
{
    //set the pos value to a negative value
    d_pos.x = -1;
    d_pos.y = -1;
    smovVectMouse.emit(PG_Point(-1,-1));
}
